@@ -6,13 +6,14 @@ from typing import Any, Awaitable, Callable, Dict, List, Optional, cast
 from anthropic import AI_PROMPT, HUMAN_PROMPT, AsyncAnthropic
 
 from llmvm.common.container import Container
-from llmvm.common.logging_helpers import messages_trace, setup_logging
-from llmvm.common.objects import (Assistant, AstNode, Content, Executor,
-                                  FileContent, ImageContent, Message,
-                                  PdfContent, System, TokenStopNode, User,
-                                  awaitable_none)
-from llmvm.common.perf import (TokenPerf, TokenStreamManager)
 from llmvm.common.helpers import Helpers
+from llmvm.common.logging_helpers import messages_trace, setup_logging
+from llmvm.common.object_transformers import ObjectTransformers
+from llmvm.common.objects import (Assistant, AstNode, Content, Executor,
+                                  FileContent, ImageContent, MarkdownContent,
+                                  Message, PdfContent, System, TokenStopNode,
+                                  User, awaitable_none)
+from llmvm.common.perf import TokenPerf, TokenStreamManager
 
 logging = setup_logging()
 
@@ -21,24 +22,21 @@ class AnthropicExecutor(Executor):
     def __init__(
         self,
         api_key: str,
-        default_model: str = 'claude-2.1',
+        default_model: str = 'claude-3-sonnet-20240229',
         api_endpoint: str = 'https://api.anthropic.com',
         default_max_token_len: int = 200000,
-        default_max_completion_len: int = 2048,
+        default_max_output_len: int = 4096,
         beta: bool = True,
     ):
-        self.default_max_token_len = default_max_token_len
-        self.default_max_completion_len = default_max_completion_len
-        self.default_model = default_model
+        super().__init__(
+            default_model=default_model,
+            api_endpoint=api_endpoint,
+            default_max_token_len=default_max_token_len,
+            default_max_output_len=default_max_output_len,
+        )
         self.client = AsyncAnthropic(api_key=api_key, base_url=api_endpoint)
         self.beta = beta
 
-    def messages_trace(self, executor_name: str, message: List[Dict[str, str]]):
-        if Container.get_config_variable('LLMVM_EXECUTOR_TRACE', default=''):
-            with open(os.path.expanduser(Container.get_config_variable('LLMVM_EXECUTOR_TRACE')), 'a+') as f:
-                for m in message:
-                    f.write(f"{m['role'].capitalize()}: {m['content']}\n\n")
-
     def from_dict(self, message: Dict[str, Any]) -> 'Message':
         role = message['role']
         message_content = message['content']
@@ -85,11 +83,14 @@ class AnthropicExecutor(Executor):
     def wrap_messages(self, messages: List[Message]) -> List[Dict[str, str]]:
         def wrap_message(index: int, content: Content) -> str:
             if isinstance(content, FileContent):
-                return f"<file url={content.url}>{content.get_content()}</file>"
+                return f"<file url={content.url}>{content.get_str()}</file>"
             elif isinstance(content, PdfContent):
-                return f"<pdf url={content.url}>{content.get_content()}</pdf>"
+                # return f"<pdf url={content.url}>{content.get_str()}</pdf>"
+                return f"<pdf url={content.url}>{ObjectTransformers.transform_pdf_content(content, self)}</pdf>"
+            elif isinstance(content, MarkdownContent):
+                return f"{ObjectTransformers.transform_markdown_content(content, self)}"
             else:
-                return f"{content.get_content()}"
+                return f"{content.get_str()}"
 
         wrapped = []
 
@@ -101,46 +102,72 @@ class AnthropicExecutor(Executor):
         system_message = ''
         if len(system_messages) > 0:
             system_message = system_messages[-1]
-            wrapped.append({'role': 'system', 'content': system_message.message.get_content()})
+            wrapped.append({'role': 'system', 'content': system_message.message.get_str()})
 
         messages = [m for m in messages if m.role() != 'system']
 
+        # check to see if there are more than 20 images in the message list
+        image_count = len([m for m in messages if isinstance(m, User) and isinstance(m.message, ImageContent)])
+        if image_count >= 20:
+            logging.debug(f'Image count is {image_count}, filtering.')
+
+            # get the top 20 ordered by size, then remove the rest
+            images = [m for m in messages if isinstance(m, User) and isinstance(m.message, ImageContent)]
+            images.sort(key=lambda x: len(x.message.sequence), reverse=True)
+            smaller_images = images[20:]
+
+            # remove the images from the messages list and collapse the previous and next message
+            # if there is no other images in between
+            for image in smaller_images:
+                index = messages.index(image)
+                if (
+                    index > 0
+                    and not isinstance(messages[index - 1].message, ImageContent)
+                    and not isinstance(messages[index + 1].message, ImageContent)
+                ):
+                    previous = messages[index - 1]
+                    next = messages[index + 1]
+                    previous.message = Content(previous.message.get_str() + next.message.get_str())
+                    messages.remove(image)
+                    messages.remove(next)
+                else:
+                    messages.remove(image)
+
         counter = 1
         for i in range(len(messages)):
             if isinstance(messages[i], User) and isinstance(messages[i].message, ImageContent):
                 # figure out
-                wrapped.append({
-                    'role': 'user',
-                    'content': [
-                        {
-                            'type': 'image',
-                            'source': {
-                                "type": "base64",
-                                "media_type": Helpers.classify_image(messages[i].message.sequence),
-                                "data": base64.b64encode(messages[i].message.sequence).decode('utf-8')  # type: ignore
-                            }
-                        }
-                    ]
-                })
+                if 'image/unknown' not in Helpers.classify_image(messages[i].message.sequence):
+                    wrapped.append({
+                        'role': 'user',
+                        'content': [{
+                                'type': 'image',
+                                'source': {
+                                    "type": "base64",
+                                    "media_type": Helpers.classify_image(messages[i].message.sequence),
+                                    "data": base64.b64encode(Helpers.anthropic_resize(messages[i].message.sequence)).decode('utf-8')  # type: ignore
+                                }
+                        }]
+                    })
             elif isinstance(messages[i], User) and i < len(messages) - 1:  # is not last message, context messages
                 wrapped.append({'role': 'user', 'content': wrap_message(counter, messages[i].message)})
                 counter += 1
             elif isinstance(messages[i], User) and i == len(messages) - 1:  # is the last message
-                wrapped.append({'role': 'user', 'content': messages[i].message.get_content()})
+                wrapped.append({'role': 'user', 'content': messages[i].message.get_str()})
             else:
-                wrapped.append({'role': messages[i].role(), 'content': messages[i].message.get_content()})
+                wrapped.append({'role': messages[i].role(), 'content': messages[i].message.get_str()})
 
         messages_list = []
 
-        for i in range(len(wrapped)):
-            if i > 0 and wrapped[i]['role'] == wrapped[i - 1]['role']:
-                if wrapped[i]['role'] == 'user':
-                    messages_list.append({'role': 'assistant', 'content': 'Thanks. I am ready for your next message.'})
-                elif wrapped[i]['role'] == 'assistant':
-                    messages_list.append({'role': 'user', 'content': 'Thanks. I am ready for your next message.'})
-            messages_list.append(wrapped[i])
+        # for i in range(len(wrapped)):
+        #     if i > 0 and wrapped[i]['role'] == wrapped[i - 1]['role']:
+        #         if wrapped[i]['role'] == 'user':
+        #             messages_list.append({'role': 'assistant', 'content': 'Thanks. Ready for next message.'})
+        #         elif wrapped[i]['role'] == 'assistant':
+        #             messages_list.append({'role': 'user', 'content': 'Thanks. Read for your next message.'})
+        #     messages_list.append(wrapped[i])
 
-        return messages_list
+        return wrapped
 
     def user_token(self):
         if self.beta:
@@ -160,91 +187,37 @@ class AnthropicExecutor(Executor):
         else:
             return AI_PROMPT
 
-    def max_tokens(self, model: Optional[str]) -> int:
-        model = model if model else self.default_model
-        match model:
-            case 'claude-2.1':
-                return 200000
-            case 'claude-2.0':
-                return 200000
-            case str(model) if model.startswith('claude-3'):
-                return 200000
-            case 'claude-instant-1.2':
-                return 100000
-            case _:
-                logging.debug('max_tokens() is not implemented for model {}, returning 200000'.format(model))
-                return 200000
-
-    def set_default_model(self, default_model: str):
-        self.default_model = default_model
-
-    def get_default_model(self):
-        return self.default_model
-
-    def set_default_max_tokens(self, default_max_tokens: int):
-        self.default_max_token_len = default_max_tokens
-
-    def max_prompt_tokens(
-        self,
-        completion_token_len: Optional[int] = None,
-        model: Optional[str] = None,
-    ) -> int:
-        return self.max_tokens(model) - (completion_token_len if completion_token_len else self.default_max_completion_len)
-
-    def max_completion_tokens(
-        self,
-        model: Optional[str] = None,
-    ):
-        return self.default_max_completion_len
-
     def count_tokens(
         self,
         messages: List[Message] | List[Dict[str, str]] | str,
         model: Optional[str] = None,
     ) -> int:
-        model_str = model if model else self.default_model
-
         async def tokenizer_len(content: str | List) -> int:
+            # image should have already been resized
             if isinstance(content, list) and len(content) > 0 and isinstance(content[0], dict) and 'source' in content[0]:
-                # todo not supported
-                return 0
-            return await self.client.count_tokens(str(content))
+                token_count = Helpers.anthropic_image_tok_count(content[0]['source']['data'])
+                return token_count
 
-        def num_tokens_from_messages(messages, model: str):
+            token_count = await self.client.count_tokens(str(content))
+            return token_count
+
+        def num_tokens_from_messages(messages):
             # this is inexact, but it's a reasonable approximation
-            if model in {
-                'claude-2',
-                'claude-2.0',
-                'claude-2.1',
-                'claude-instant-1.2',
-                'claude-instant-1',
-            }:
-                tokens_per_message = 3
-                tokens_per_name = 1
-            elif model.startswith('claude-3'):
-                tokens_per_message = 3
-                tokens_per_name = 1
-            else:
-                logging.debug(f"num_tokens_from_messages() is not implemented for model {model}.")
-                tokens_per_message = 3
-                tokens_per_name = 1
             num_tokens = 0
+            tokens_per_message = 4
             for message in messages:
                 num_tokens += tokens_per_message
-                for key, value in message.items():
+                for _, value in message.items():
                     num_tokens += asyncio.run(tokenizer_len(value))
-                    if key == "name":
-                        num_tokens += tokens_per_name
-            num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
             return num_tokens
 
         if isinstance(messages, list) and len(messages) > 0 and isinstance(messages[0], Message):
             dict_messages = self.wrap_messages(cast(List[Message], messages))
-            return num_tokens_from_messages(dict_messages, model=model_str)
+            return num_tokens_from_messages(dict_messages)
         elif isinstance(messages, list) and len(messages) > 0 and isinstance(messages[0], dict):
-            return num_tokens_from_messages(messages, model=model_str)
+            return num_tokens_from_messages(messages)
         elif isinstance(messages, str):
-            return num_tokens_from_messages(self.wrap_messages([User(Content(messages))]), model=model_str)
+            return num_tokens_from_messages(self.wrap_messages([User(Content(messages))]))
         else:
             raise ValueError('cannot calculate tokens for messages: {}'.format(messages))
 
@@ -267,7 +240,7 @@ class AnthropicExecutor(Executor):
         messages: List[Dict[str, Any]],
         functions: List[Dict[str, str]] = [],
         model: Optional[str] = None,
-        max_completion_tokens: int = 2048,
+        max_output_tokens: int = 2048,
         temperature: float = 0.0,
     ) -> TokenStreamManager:
         model = model if model else self.default_model
@@ -276,11 +249,11 @@ class AnthropicExecutor(Executor):
             raise NotImplementedError('functions are not implemented for ClaudeExecutor')
 
         message_tokens = self.count_tokens(messages=messages, model=model)
-        if message_tokens > self.max_prompt_tokens(max_completion_tokens, model=model):
-            raise Exception('Prompt too long. prompt tokens: {}, completion tokens: {}, total: {}, max context window: {}'
+        if message_tokens > self.max_input_tokens(max_output_tokens, model=model):
+            raise Exception('Prompt too long. input tokens: {}, output tokens: {}, total: {}, max sontext window: {}'
                             .format(message_tokens,
-                                    max_completion_tokens,
-                                    message_tokens + max_completion_tokens,
+                                    max_output_tokens,
+                                    message_tokens + max_output_tokens,
                                     self.max_tokens(model)))
 
         # the messages API does not accept System messages, only User and Assistant messages.
@@ -294,8 +267,25 @@ class AnthropicExecutor(Executor):
         # anthropic disallows empty messages, so we're going to remove any Message that doesn't contain content
         for message in messages:
             if not message['content'] or message['content'] == '' or message['content'] == b'':
+                logging.warning(f"Removing empty message: {message}")
                 messages.remove(message)
 
+        if Container().get_config_variable('ANTHROPIC_COLLAPSE_MESSAGES', default=False):
+            collapsed_messages = []
+            accumulator = ''
+            for i in range(len(messages)):
+                if messages[i]['role'] == 'user' and not isinstance(messages[i]['content'], list):
+                    accumulator += messages[i]['content'] + '\n\n'
+                elif messages[i]['role'] == 'user' and isinstance(messages[i]['content'], list) and accumulator:
+                    collapsed_messages.append({'role': 'user', 'content': accumulator})
+                    collapsed_messages.append(messages[i])
+                    accumulator = ''
+                else:
+                    collapsed_messages.append(messages[i])
+            if accumulator:
+                collapsed_messages.append({'role': 'user', 'content': accumulator})
+            messages = collapsed_messages
+
         # the messages API also doesn't allow for multiple User or Assistant messages in a row, so we're
         # going to add an Assistant message in between two User messages, and a User message between two Assistant.
         messages_list: List[Dict[str, str]] = []
@@ -308,7 +298,12 @@ class AnthropicExecutor(Executor):
                     messages_list.append({'role': 'user', 'content': 'Thanks. I am ready for your next message.'})
             messages_list.append(messages[i])
 
-        self.messages_trace(self.name(), messages_list)
+        # todo, this is a busted hack. if a helper function returns nothing, then usually that
+        # message get stripped away
+        if messages_list[0]['role'] != 'system' and messages_list[0]['role'] != 'user':
+            messages_list.insert(0, {'role': 'user', 'content': 'None.'})
+
+        messages_trace(messages_list)
 
         token_trace = TokenPerf('__aexecute_direct', 'openai', model, prompt_len=message_tokens)  # type: ignore
         token_trace.start()
@@ -317,30 +312,30 @@ class AnthropicExecutor(Executor):
             if self.beta:
                 # AsyncStreamManager[AsyncMessageStream]
                 stream = self.client.messages.stream(
-                    max_tokens=max_completion_tokens,
+                    max_tokens=max_output_tokens,
                     messages=messages_list,  # type: ignore
                     model=model,
                     system=system_message,
-                    temperature=0.0,
+                    temperature=temperature,
                 )
                 return TokenStreamManager(stream, token_trace)
             else:
                 stream = await self.client.completions.create(
-                    max_tokens_to_sample=max_completion_tokens,
+                    max_tokens_to_sample=max_output_tokens,
                     model=model,
                     stream=True,
-                    temperature=0.0,
+                    temperature=temperature,
                     prompt=self.__format_prompt(messages_list),
                 )
                 return TokenStreamManager(stream, token_trace)
         except Exception as e:
-            print(e)
+            logging.error(e)
             raise
 
     async def aexecute(
         self,
         messages: List[Message],
-        max_completion_tokens: int = 2048,
+        max_output_tokens: int = 2048,
         temperature: float = 0.0,
         model: Optional[str] = None,
         stream_handler: Callable[[AstNode], Awaitable[None]] = awaitable_none,
@@ -369,7 +364,7 @@ class AnthropicExecutor(Executor):
 
         stream = self.__aexecute_direct(
             messages_list,
-            max_completion_tokens=max_completion_tokens,
+            max_output_tokens=max_output_tokens,
             model=model,
             temperature=temperature,
         )
@@ -396,14 +391,12 @@ class AnthropicExecutor(Executor):
         messages_list.append({'role': 'assistant', 'content': text_response})
         conversation: List[Message] = [self.from_dict(m) for m in messages_list]
 
-        messages_trace(messages_list)
-
         assistant = Assistant(
             message=conversation[-1].message,
             messages_context=conversation
         )
         assistant.perf_trace = perf
-        if assistant.message.get_content() == '':
+        if assistant.message.get_str() == '':
             logging.error(f'Assistant message is empty. Returning empty message. {perf.request_id or ""}')
 
         return assistant
@@ -411,7 +404,7 @@ class AnthropicExecutor(Executor):
     def execute(
         self,
         messages: List[Message],
-        max_completion_tokens: int = 2048,
+        max_output_tokens: int = 2048,
         temperature: float = 0.0,
         model: Optional[str] = None,
         stream_handler: Optional[Callable[[AstNode], None]] = None,
@@ -420,4 +413,4 @@ class AnthropicExecutor(Executor):
             if stream_handler:
                 stream_handler(node)
 
-        return asyncio.run(self.aexecute(messages, max_completion_tokens, temperature, model, stream_pipe))
+        return asyncio.run(self.aexecute(messages, max_output_tokens, temperature, model, stream_pipe))
