@@ -1,8 +1,8 @@
 import base64
 import copy
 import datetime as dt
-import functools
 import importlib
+import json
 import os
 from abc import ABC, abstractmethod
 from enum import Enum
@@ -14,6 +14,65 @@ from pydantic import BaseModel
 T = TypeVar('T')
 
 
+class TokenPriceCalculator():
+    def __init__(
+        self,
+        price_file: str = 'model_prices_and_context_window.json',
+    ):
+        self.price_file = resources.files('llmvm') / price_file
+        self.prices = self.__load_prices()
+
+    def __load_prices(self):
+        with open(self.price_file, 'r') as f:  # type: ignore
+            json_prices = json.load(f)
+            return json_prices
+
+    def get(self, model: str, key: str, executor: Optional[str] = None) -> Optional[Any]:
+        if model in self.prices and key in self.prices[model]:
+            return self.prices[model][key]
+        elif executor and f'{executor}/{model}' in self.prices and key in self.prices[f'{executor}/{model}']:
+            return self.prices[f'{executor}/{model}'][key]
+        return None
+
+    def input_price(
+        self,
+        model: str,
+        executor: Optional[str] = None
+    ) -> float:
+        return self.get(model, 'input_cost_per_token', executor) or 0.0
+
+    def output_price(
+        self,
+        model: str,
+        executor: Optional[str] = None
+    ) -> float:
+        return self.get(model, 'output_cost_per_token', executor) or 0.0
+
+    def max_tokens(
+        self,
+        model: str,
+        executor: Optional[str] = None,
+        default: int = 0
+    ) -> int:
+        return self.get(model, 'max_tokens', executor) or default
+
+    def max_input_tokens(
+        self,
+        model: str,
+        executor: Optional[str] = None,
+        default: int = 0
+    ) -> int:
+        return self.get(model, 'max_input_tokens', executor) or default
+
+    def max_output_tokens(
+        self,
+        model: str,
+        executor: Optional[str] = None,
+        default: int = 0
+    ) -> int:
+        return self.get(model, 'max_output_tokens', executor) or default
+
+
 def bcl(module_or_path):
     def decorator(cls):
         class NewClass(cls):
@@ -58,11 +117,23 @@ class Visitor(ABC):
 
 
 class Executor(ABC):
+    def __init__(
+        self,
+        default_model: str,
+        api_endpoint: str,
+        default_max_token_len: int,
+        default_max_output_len: int,
+    ):
+        self.default_model = default_model
+        self.api_endpoint = api_endpoint
+        self.default_max_token_len = default_max_token_len
+        self.default_max_output_len = default_max_output_len
+
     @abstractmethod
     async def aexecute(
         self,
         messages: List['Message'],
-        max_completion_tokens: int = 2048,
+        max_output_tokens: int = 2048,
         temperature: float = 1.0,
         stream_handler: Optional[Callable[['AstNode'], Awaitable[None]]] = None,
         model: Optional[str] = None,
@@ -70,11 +141,50 @@ class Executor(ABC):
     ) -> 'Assistant':
         pass
 
+    def set_default_max_tokens(
+        self,
+        default_max_token_len: int,
+    ) -> None:
+        self.default_max_token_len = default_max_token_len
+
+    def set_default_model(
+        self,
+        default_model: str,
+    ) -> None:
+        self.default_model = default_model
+
+    def get_default_model(
+        self,
+    ) -> str:
+        return self.default_model
+
+    def max_tokens(self, model: Optional[str]) -> int:
+        return TokenPriceCalculator().max_tokens(model or self.default_model, default=self.default_max_token_len)
+
+    def max_input_tokens(
+        self,
+        output_token_len: Optional[int] = None,
+        model: Optional[str] = None,
+    ) -> int:
+        return TokenPriceCalculator().max_input_tokens(
+            model or self.default_model,
+            default=self.default_max_token_len - self.default_max_output_len
+        )
+
+    def max_output_tokens(
+        self,
+        model: Optional[str] = None,
+    ) -> int:
+        return TokenPriceCalculator().max_output_tokens(
+            model or self.default_model,
+            default=self.default_max_output_len
+        )
+
     @abstractmethod
     def execute(
         self,
         messages: List['Message'],
-        max_completion_tokens: int = 2048,
+        max_output_tokens: int = 2048,
         temperature: float = 1.0,
         stream_handler: Optional[Callable[['AstNode'], None]] = None,
         model: Optional[str] = None,
@@ -82,49 +192,10 @@ class Executor(ABC):
     ) -> 'Assistant':
         pass
 
-    @abstractmethod
-    def set_default_max_tokens(
-        self,
-        default_max_tokens: int,
-    ):
-        pass
-
-    @abstractmethod
-    def set_default_model(
-        self,
-        default_model: str,
-    ):
-        pass
-
-    @abstractmethod
-    def get_default_model(
-        self,
-    ) -> str:
-        pass
-
     @abstractmethod
     def name(self) -> str:
         pass
 
-    @abstractmethod
-    def max_tokens(self, model: Optional[str]) -> int:
-        pass
-
-    @abstractmethod
-    def max_prompt_tokens(
-        self,
-        completion_token_len: Optional[int] = None,
-        model: Optional[str] = None,
-    ) -> int:
-        pass
-
-    @abstractmethod
-    def max_completion_tokens(
-        self,
-        model: Optional[str] = None,
-    ) -> int:
-        pass
-
     @abstractmethod
     def count_tokens(
         self,
@@ -351,7 +422,7 @@ class DebugNode(AstNode):
 class Content(AstNode):
     def __init__(
         self,
-        sequence: Optional[AstNode | List[AstNode] | str | bytes | Any] = None,
+        sequence: Optional[AstNode | List[AstNode] | List['Content'] | str | bytes | Any] = None,
         content_type: str = 'text',
         url: str = '',
     ):
@@ -361,6 +432,7 @@ class Content(AstNode):
 
         self.content_type = content_type
         self.url = url
+        self.original_sequence: object = None
 
         if isinstance(sequence, str):
             self.sequence = [sequence]
@@ -368,6 +440,8 @@ class Content(AstNode):
             self.sequence = sequence
         elif isinstance(sequence, Content):
             self.sequence = sequence.sequence  # type: ignore
+        elif isinstance(sequence, list) and len(sequence) > 0 and isinstance(sequence[0], Content):
+            self.sequence = sequence
         elif isinstance(sequence, AstNode):
             self.sequence = [sequence]
         elif isinstance(sequence, list) and len(sequence) > 0 and isinstance(sequence[0], AstNode):
@@ -393,7 +467,7 @@ class Content(AstNode):
     def __repr__(self):
         return f'Content({self.sequence})'
 
-    def get_content(self) -> str:
+    def get_str(self) -> str:
         return self.__str__()
 
     def b64encode(self) -> str:
@@ -401,6 +475,8 @@ class Content(AstNode):
             return base64.b64encode(self.sequence).decode('utf-8')
         elif isinstance(self.sequence, str):
             return base64.b64encode(self.sequence.encode('utf-8')).decode('utf-8')
+        elif isinstance(self.sequence, list) and len(self.sequence) > 0 and isinstance(self.sequence[0], Content):
+            return base64.b64encode(self.original).decode('utf-8')
         else:
             raise ValueError(f'unknown sequence: {self.sequence}')
 
@@ -418,11 +494,33 @@ class ImageContent(Content):
         super().__init__(sequence, 'image', url)
         self.sequence = sequence
 
+    def __str__(self):
+        return f'ImageContent({self.url})'
+
+    def __repr__(self):
+        return f'ImageContent({self.url})'
+
+
+class MarkdownContent(Content):
+    def __init__(
+        self,
+        sequence: str | List[Content],
+        url: str = '',
+    ):
+        super().__init__(sequence, 'markdown', url)
+        self.sequence = sequence
+
+    def __str__(self):
+        return f'MarkdownContent({self.url})'
+
+    def get_str(self) -> str:
+        return str(self.sequence)
+
 
 class PdfContent(Content):
     def __init__(
         self,
-        sequence: bytes | str,
+        sequence: bytes | List[Content],
         url: str = '',
     ):
         super().__init__(sequence, 'pdf', url)
@@ -434,48 +532,26 @@ class PdfContent(Content):
     def is_local(self):
         return os.path.isfile(self.url)
 
-    def is_text(self):
-        return isinstance(self.sequence, str)
-
-    def get_content(self) -> str:
-        def late_bind_pdf_parse(obj, method_name: str):
-            # hoooly crap this is a hack.
-            module = importlib.import_module('llmvm.server.tools.pdf')
-            cls = getattr(module, 'PdfHelpers')
-            method = getattr(cls, method_name)
-            return method(obj)
-
-        if self.is_text():
-            return str(self.sequence)
-        elif self.is_local():
-            self.sequence = late_bind_pdf_parse(self.url, 'parse_pdf')
-            return self.sequence
-        elif self.url:
-            self.sequence = late_bind_pdf_parse(self.url, 'parse_pdf')
-            return self.sequence
-        elif isinstance(self.sequence, bytes):
-            self.sequence = late_bind_pdf_parse(self.sequence, 'parse_pdf_bytes')
-            return self.sequence
-        else:
-            raise ValueError('cannot get text from pdf')
+    def get_str(self) -> str:
+        raise NotImplementedError('PdfContent.get_str() not implemented')
 
 
 class FileContent(Content):
     def __init__(
         self,
-        sequence: bytes,
+        sequence: bytes | List[Content],
         url: str = '',
     ):
         super().__init__(sequence, 'file', url)
         self.sequence = sequence
 
-    def is_local(self):
-        return os.path.isfile(self.url)
-
     def __str__(self):
         return f'FileContent({self.url})'
 
-    def get_content(self):
+    def is_local(self):
+        return os.path.isfile(self.url)
+
+    def get_str(self):
         if self.is_local():
             with open(self.url, 'r') as f:
                 return f.read()
@@ -557,7 +633,7 @@ class Message(AstNode):
     @staticmethod
     def to_dict(message: 'Message', server_serialization: bool = False) -> Dict[str, Any]:
         def file_wrap(message: FileContent | PdfContent):
-            return f'The following data/content is from this url: {message.url}\n\n{message.get_content()}'
+            return f'The following data/content is from this url: {message.url}\n\n{message.get_str()}'
 
         # primarily to pass to Anthropic or OpenAI api
         if isinstance(message, User) and isinstance(message.message, ImageContent):
