@@ -38,6 +38,23 @@ nest_asyncio.apply()
 
 logging = setup_logging()
 
+# check to see if the config file exists, if not, create it
+try:
+    Container().get_config_variable('executor', 'LLMVM_EXECUTOR', default='')
+except ValueError:
+    rich.print('[cyan]Configuration file not found. Adding default config in ~/.config/llmvm/config.yaml[/cyan]')
+    os.makedirs(os.path.expanduser('~/.config/llmvm'), exist_ok=True)
+    os.makedirs(os.path.expanduser('~/.local/share/llmvm'), exist_ok=True)
+    os.makedirs(os.path.expanduser('~/.local/share/llmvm/cache'), exist_ok=True)
+    os.makedirs(os.path.expanduser('~/.local/share/llmvm/download'), exist_ok=True)
+    os.makedirs(os.path.expanduser('~/.local/share/llmvm/cdn'), exist_ok=True)
+    os.makedirs(os.path.expanduser('~/.local/share/llmvm/logs'), exist_ok=True)
+    os.makedirs(os.path.expanduser('~/.local/share/llmvm/faiss'), exist_ok=True)
+
+    config_file = resources.files('llmvm') / 'config.yaml'
+    shutil.copy(str(config_file), os.path.expanduser('~/.config/llmvm/config.yaml'))
+
+
 app = FastAPI()
 
 agents = list(
@@ -163,7 +180,8 @@ async def stream_response(response):
 @app.post('/v1/chat/completions')
 async def chat_completions(request: Request):
     api_key = os.environ.get('OPENAI_API_KEY', default=''),
-    aclient = AsyncOpenAI(api_key=cast(str,api_key))
+    aclient = AsyncOpenAI(api_key=cast(str, api_key))
+    client = OpenAI(api_key=cast(str, api_key))
 
     try:
         # Construct the prompt from the messages
@@ -238,6 +256,7 @@ async def download(
         thread.id = temp.id
 
     queue = asyncio.Queue()
+    controller = get_controller()
 
     async def callback(token: AstNode):
         queue.put_nowait(token)
@@ -245,25 +264,21 @@ async def download(
     async def stream():
         async def execute_and_signal():
             stream_handler = callback
-            from bcl import ContentDownloader
+            from llmvm.server.base_library.content_downloader import \
+                ContentDownloader
 
             # todo thread cookies through here
             downloader = ContentDownloader(
                 expr=download_item.url,
-                agents=[],
-                messages=[],
-                starlark_runtime=get_controller().starlark_runtime,
-                original_code='',
-                original_query=''
             )
-            content = downloader.download()
+            content: Content = downloader.download()
             queue.put_nowait(StopNode())
 
             if content:
                 background_tasks.add_task(
                     vector_search.ingest_text,
-                    content,
-                    content[:25],
+                    controller.statement_to_str(content),
+                    controller.statement_to_str(content)[:25],
                     download_item.url,
                     {}
                 )
@@ -465,7 +480,7 @@ async def code_completions(request: SessionThread):
 async def tools_completions(request: SessionThread):
     thread = request
 
-    if not cache_session.has_key(thread.id) or thread.id <= 0:
+    if not cache_session.has_key(thread.id) or thread.id == 0:
         temp = __get_thread(0)
         thread.id = temp.id
 
@@ -488,7 +503,8 @@ async def tools_completions(request: SessionThread):
         thread.executor = controller.get_executor().name()
         thread.model = model
 
-    logging.debug(f'/v1/chat/tools_completions?id={thread.id}&mode={mode}&model={model}&executor={thread.executor}&compression={thread.compression}&cookies={thread.cookies}')
+    logging.debug(f'/v1/chat/tools_completions?id={thread.id}&mode={mode}&model={model} \
+                  &executor={thread.executor}&compression={thread.compression}&cookies={thread.cookies}')
 
     if len(messages) == 0:
         raise HTTPException(status_code=400, detail='No messages provided')
@@ -570,26 +586,12 @@ async def general_exception_handler(request: Request, exc: Exception):
 
 
 if __name__ == '__main__':
-    try:
-        Container().get_config_variable('executor', 'LLMVM_EXECUTOR', default='')
-    except ValueError:
-        rich.print('[cyan]Configuration file not found. Adding default config in ~/.config/llmvm/config.yaml[/cyan]')
-        os.makedirs(os.path.expanduser('~/.config/llmvm'), exist_ok=True)
-        os.makedirs(os.path.expanduser('~/.local/share/llmvm'), exist_ok=True)
-        os.makedirs(os.path.expanduser('~/.local/share/llmvm/cache'), exist_ok=True)
-        os.makedirs(os.path.expanduser('~/.local/share/llmvm/download'), exist_ok=True)
-        os.makedirs(os.path.expanduser('~/.local/share/llmvm/cdn'), exist_ok=True)
-        os.makedirs(os.path.expanduser('~/.local/share/llmvm/logs'), exist_ok=True)
-        os.makedirs(os.path.expanduser('~/.local/share/llmvm/faiss'), exist_ok=True)
-
-        config_file = resources.files('llmvm') / 'config.yaml'
-        shutil.copy(str(config_file), os.path.expanduser('~/.config/llmvm/config.yaml'))
-
     default_controller = Container().get_config_variable('executor', 'LLMVM_EXECUTOR', default='')
     default_model_str = f'{default_controller}_model'
     default_model = Container().get_config_variable(default_model_str, 'LLMVM_MODEL', default='')
     rich.print(f'[cyan]Default executor is: {default_controller}[/cyan]')
     rich.print(f'[cyan]Default model is: {default_model}[/cyan]')
+    rich.print(f'[cyan]Make sure to `playwright install firefox`.[/cyan]')
 
     for agent in agents:
         rich.print(f'[green]Loaded agent: {agent.__name__}[/green]')  # type: ignore
